from yarok.components.anet_a30.anet_at30 import AnetA30
from yarok.components.cam.cam import Cam
from yarok.mjc.interface import InterfaceMJC
from yarok.components_manager import component

from yarok.components.robotiq_2f85.robotiq_2f85 import robotiq_2f85
from yarok.components.ur5.ur5 import UR5
from yarok.components.gelsight2014.gelsight2014 import gelsight2014
from yarok.components.geltip.geltip import geltip


class EmptyWorldInterfaceMJC:
    def __init__(self, interface: InterfaceMJC):
        self.interface = interface

    def step(self):
        # frame = self.interface.get_frame('extrinsic_cam')
        # print(frame.shape)
        # cv2.imshow('frame', frame)
        # cv2.waitKey(1)
        pass

@component(
    components=[
        UR5,
        robotiq_2f85,
        gelsight2014,
        geltip,
        AnetA30,
        Cam
    ],
    interface_mjc=EmptyWorldInterfaceMJC
)
class EmptyWorld:

    def __init__(self):
        # self.mani = mani
        # print('===> ', mani)
        # pass
        pass

    def init(self):
        # self.mani.move([pi / 2, 0, 0, pi / 2, pi / 2, pi / 2])
        pass


class InterfaceMJC:

    def __init__(self, name, sim, model):
        self.component_name = name
        self.sim = sim
        self.model = model
        self.sensor_name2id = {}
        self.actuator_name2id = {}
        self.camera_name2id = {}
        self.actuators = []
        self.sensors = []
        self.cameras = []
        self.component_name = ''

    def set_ctrl(self, k, v):
        actuator_name = self.actuators[k] if type(k) == int else k
        self.sim.data.ctrl[self.actuator_name2id[actuator_name]] = v


    def sensordata(self):
        return [self.sim.data.sensordata[self.sensor_name2id[sn]] for sn in self.sensors]


    def get_frame(self, camera_name, shape=(480, 640), depth=False):
        return self.sim.render(height=shape[0],
                               width=shape[1],
                               camera_name=self.component_name + ':' + camera_name,
                               depth=depth,
                               mode='offscreen')



from yarok.components_manager import component
from yarok.mjc.interface import InterfaceMJC


class CameraInterfaceMJC:

    def __init__(self, interface: InterfaceMJC):
        self.interface = interface

    def get_frame(self):
        return self.interface.get_frame('cam')


@component(
    interface_mjc=CameraInterfaceMJC
)
class Cam:

    def __init__(self):
        pass



from yarok.components_manager import component
from yarok.mjc.interface import InterfaceMJC


class AnetA30InterfaceMJC:

    def __init__(self, interface: InterfaceMJC):
        self.interface = interface
        self.gear = 100
        self.target_position = (0, 0, 0)
        self.MAX_X = 320
        self.MAX_Y = 320
        self.MAX_Z = 420

    def move(self, p):
        self.target_position = [
            max(0, min(p[0], self.MAX_X)),
            max(0, min(p[1], self.MAX_Y)),
            max(0, min(p[2], self.MAX_Z))
        ]

    def step(self):
        x, y, z = self.target_position

        self.interface.set_ctrl('ax', (x / 1000.0) * self.gear)
        self.interface.set_ctrl('ay', (y / 1000.0) * self.gear)
        self.interface.set_ctrl('az', (z / 1000.0) * self.gear)


@component(
    interface_mjc=AnetA30InterfaceMJC
)
class AnetA30:

    def __init__(self):
        pass

    def move(self):
        pass



from .components_manager import ComponentsManager

from .mjc.platform import PlatformMJC


def load(world, config):
    def get_config(key, default):
        return config[key] if key in config else default

    manager = ComponentsManager(
        root=world,
    )

    platform = get_config('plaform', PlatformMJC)(manager, **get_config('platform_args', {}))

    manager.init_components(platform.interfaces)

    return platform, manager


def run(world, config):
    platform, _ = load(world, config)

    while True:
        platform.step()



import inspect
import os
import pathlib
import re
import sys
import xml.etree.ElementTree as ET
from xml.etree.ElementTree import tostring

from mujoco_py import load_model_from_xml, MjSim


class ComponentsManager:
    """
        Handles loading components templates (.xml) & controllers (.py),
        Builds world model/components tree
    """

    def __init__(self, **kwargs):
        self.id_counter = 0
        root_component_factory = kwargs['root']
        root_component_factories = root_component_factory.__data__['components']

        self.root = {
            'factory': root_component_factory,
            'name': '',
            'id': self.gen_name(),
            'children': []
        }

        self.xml_tree = self.load_tree(self.root, components_factories=root_component_factories)

        self.components = {}

        # build components map and init interfaces
        def walk_components(component):
            for child in component['children']:
                walk_components(child)

            self.components[component['id']] = component

        walk_components(self.root)

        # root['instance'].init()

        # def init(self, sim):
        #     self.init_interfaces(self.root_component, self.model, sim)
        #     self.init_components(root_component)

    def get(self, n):
        return self.components[n]

    def data(self, n):
        return self.components[n]['factory'].__data__

    # def step(self):
    # def step_components(component):
    #     for child in component['children']:
    #         step_components(child)
    #
    #     if 'interface_instance' in component['factory'].__data__ \
    #             and hasattr(component['factory'].__data__['interface_instance'], 'step'):
    #         component['interface_instance'].step()
    #
    # step_components(self.root)

    # if self.viewer is not None:
    #     self.viewer.step()

    def load_tree(self, comp, sub_trees=None, components_factories=None):
        sub_trees = [] if sub_trees is None else sub_trees
        xml_path = comp['factory'].__data__['template_path']
        name = comp['name']
        id = comp['id']

        tree = ET.parse(xml_path).getroot()

        wb = tree.find('worldbody')

        # nesting all descendant content
        for sub_tree in sub_trees:
            if 'parent' in sub_tree.attrib:
                parentName = sub_tree.attrib['parent']
                parent = wb.find(".//body[@name='" + parentName + "']")
                if parent is None:
                    raise KeyError('Failed to inject ' +
                                   tostring(sub_tree, encoding="unicode").replace('\n', '').strip() +
                                   ' body[name=' +
                                   parentName +
                                   '] not being found at ' + xml_path)
                sub_tree.attrib.pop('parent')
                parent.append(sub_tree)
            else:
                wb.append(sub_tree)

        # renaming "foreign keys" to from xxx to id:xxx
        rename_attrs = ['name', 'joint', 'actuator', 'body1', 'body2', 'material', 'mesh', 'texture']

        def rename_fk(els):
            for el in els:
                if el.tag not in components_factories:
                    for attr in rename_attrs:
                        if attr in el.attrib:
                            el.attrib[attr] = id + ':' + el.attrib[attr]
                    rename_fk(list(el))

        rename_fk(list(tree))

        # repathing all local file refs to be relative to the .xml file
        dir_path = str(pathlib.Path(xml_path).parent.resolve())
        for el in tree.iter():
            if 'file' in el.attrib and not el.attrib['file'].startswith('/'):
                el.attrib['file'] = dir_path + '/' + el.attrib['file']

        blocks_names = ['asset', 'sensor', 'actuator', 'equality']
        blocks = {block_name: tree.find(block_name) for block_name in blocks_names}

        parent_map = {c: p for p in tree.iter() for c in p if c.tag in components_factories}

        rename_attrs = ['name', 'joint', 'actuator', 'body1', 'body2', 'material', 'mesh']

        # walking current file .xml tree and looking for components and instantiating them
        def walk_tree(component_trees):
            for component_tree in component_trees:
                tag = component_tree.tag
                if tag in components_factories:

                    component_type = tag  # component.attrib['type']
                    component_name = component_tree.attrib['name']
                    if component_name == '':
                        raise KeyError(
                            'No component should be named ""!, ' + component_type + '[name=root] at' + comp['name'])

                    # creates new sub component data wrapper
                    sub_component = {
                        'id': self.gen_name(component_name),
                        'name': component_name,
                        'factory': components_factories[component_type],
                        'children': []
                    }

                    # appends the newly found sub component to the previous/parent component
                    comp['children'].append(sub_component)

                    # load the new subcomponent tree
                    sub_component_tree = self.load_tree(
                        sub_component,
                        sub_trees=list(component_tree),
                        components_factories=components_factories
                    )

                    # merge all blocks actuators, sensors, etc to with the current component tree
                    for block_name in blocks_names:
                        block = sub_component_tree.find(block_name)
                        if block is not None:
                            if blocks[block_name] is not None:
                                blocks[block_name].extend(block)
                            else:
                                tree.append(block)

                    # merge the new subcomponent tree/body with the current tree
                    # and remove the "component" tag from the current tree
                    comp_wb = sub_component_tree.find('worldbody')
                    parent_map[component_tree].extend(comp_wb)
                    parent_map[component_tree].remove(component_tree)

                else:
                    walk_tree(list(component_tree))

        walk_tree(list(tree))

        return tree

    def gen_name(self, name=''):
        self.id_counter += 1
        return '__c__' + str(self.id_counter) + '-' + name

    def init_components(self, interfaces):

        # sub_components = []
        # for child in comp['children']:
        #     sub_components.extend(self.init_components(child))

        for n, comp in self.components.items():

            # sub_components = list(chain.from_iterable(sub_components))
            # sub_components_map = {t[0]: t[1] for t in sub_components}
            c_data = comp['factory'].__data__
            print(comp['name'], c_data['annotations'], self.components.keys())
            kwargs = {n: self.components[n] for n in c_data['annotations']}

            try:
                c = comp['factory'](**kwargs)
                comp['instance'] = c

                if n in interfaces:
                    # interface_factory = c_data['interfaces'][self.env]
                    # interface = interface_factory(comp['interface_instance_mjc'])
                    c_data['interface_instance'] = interfaces[n]

                    methods = [i for i in dir(interfaces[n].__class__) if i[0:2] != '__']
                    # print(methods, getattr(interface, 'on'))
                    [setattr(c, m, getattr(interfaces[n], m)) for m in methods if hasattr(c, m)]
                    # print(methods)
                    # print()

                # sub_components.append((comp['factory'], c))
            except TypeError as e:
                raise TypeError(str(e) + ', while initiating ' + str(comp['factory']) + (
                    '/' + comp['name'] if comp['name'] is not None else ''))

        # return sub_components

    # def init_interfaces(self, root_component, sim, model):
    #
    #     components = {}
    #
    #     # build components map and init interfaces
    #     def walk_components(component):
    #         component['interface_instance_mjc'] = InterfaceMJC(component['name'], sim, model)
    #         components[component['name']] = component
    #
    #         for child in component['children']:
    #             walk_components(child)
    #
    #     walk_components(root_component)
    #
    #     items = ['sensor', 'actuator', 'camera']
    #     for item in items:
    #         for idx in range(len(getattr(model, item + '_names'))):
    #             # print(model.camera_id2name(0), idx)
    #             abs_name = getattr(model, item + '_id2name')(idx)
    #             col_idx = abs_name.index(':')
    #             component_name = abs_name[:col_idx]
    #             item_name = abs_name[col_idx + 1:]
    #
    #             mjc_interface = components[component_name]['interface_instance_mjc']
    #             getattr(mjc_interface, item + '_name2id')[item_name] = idx
    #             getattr(mjc_interface, item + 's').append(item_name)
    #             mjc_interface.component_name = component_name


# CamelCase to snake_case
# https://stackoverflow.com/questions/1175208/elegant-python-function-to-convert-camelcase-to-snake-case
r1 = re.compile(r'(.)([A-Z][a-z]+)')
r2 = re.compile(r'__([A-Z])')
r3 = re.compile(r'([a-z0-9])([A-Z])')


def to_snake_case(name):
    name = r1.sub(r'\1_\2', name)
    name = r2.sub(r'_\1', name)
    name = r3.sub(r'\1_\2', name)
    return name.lower()


def component(**kwargs):
    def _(c):
        if not hasattr(c, '__data__'):
            class_name = to_snake_case(c.__name__)
            class_file_path = os.path.abspath((sys.modules[c.__module__].__file__))
            class_dir_path = str(pathlib.Path(class_file_path).parent.resolve())

            class_members = {t[0]: t[1] for t in inspect.getmembers(c)}
            init_members = {t[0]: t[1] for t in inspect.getmembers(class_members['__init__'])}
            init_annotations = init_members['__annotations__'] if '__annotations__' in init_members else {}

            interfaces = {k[10:]: kwargs[k] for k in kwargs if k.startswith('interface_')}

            # components = kwargs['components'] if 'components' in kwargs else []
            components = kwargs['components'] if 'components' in kwargs else []
            components = {c.__data__['name']: c for c in components}

            setattr(c, '__data__', {
                'name': class_name,
                'class_file_path': class_file_path,
                'class_dir_path': class_dir_path,
                'template_path': class_dir_path + '/' + class_name + '.xml',
                'components': components,
                'annotations': init_annotations,
                'interfaces': interfaces
            })
        return c

    return _


def on(**kwargs):
    def _(m):
        return m

    return _



from math import pi

from yarok.utils.PID import PID
from yarok.components_manager import component


class UR5InterfaceMJC:
    def __init__(self, interface):
        self.interface = interface
        initial_q = [
            pi / 2,
            0,
            0,
            0,
            0,
            0
        ]
        self.gear = 100
        P = 100
        I = 0.001
        D = 0.001
        self.PIDs = [PID(P, I, D) for a in interface.actuators]

    def move_q(self, q):
        # print(q)
        [self.PIDs[qa].setTarget(q[qa]) for qa in range(len(q))]
        # pass

    def step(self):
        data = self.interface.sensordata()
        [self.PIDs[a].update(data[a]) for a in range(len(data))]
        [self.interface.set_ctrl(a, self.PIDs[a].output) for a in range(len(self.interface.actuators))]


@component(
    interface_mjc=UR5InterfaceMJC,
)
class UR5:

    def __init__(self):
        pass

    def move_q(self, qs):
        pass

    def move(self, q):
        self.move_q(q)

    def move_xyz(self, xyz):
        # solve ik, call ikfast from Andy Zeng
        q = None
        self.move(q)
        pass


import cv2
from mujoco_py import MjViewerBasic, MjViewer


class ViewerMJC:

    def __init__(self, sim, mode):
        self.sim = sim
        self.mode = mode
        self.frame_counter = 0
        print('MODE!!!', mode)
        if self.mode == 'view':
            self.mjViewer = MjViewer(self.sim)
        # if self.mode == 'run':
        # self.sim.step()
        # self.frame_counter += 1
        # if self.frame_counter >= 30:
        # g_rgb = cv2.flip(g_rgb, 1)  # 1, horizontal flip
        # c_rgb = cv2.flip(c_rgb, 1)  # 1, horizontal flip
        #
        # g_rgb = cv2.cvtColor(g_rgb, cv2.COLOR_RGB2BGR)
        # c_rgb = cv2.cvtColor(c_rgb, cv2.COLOR_RGB2BGR)
        #
        # cv2.imshow('geltip', g_rgb)
        # cv2.imshow('global', c_rgb)

        # else:

    def step(self):
        if self.mode == 'run':
                                    # camera_name='global',
            c_rgb = self.sim.render(width=1000,
                                    height=1000,
                                    depth=False,
                                    mode='offscreen')
            c_bgr = cv2.cvtColor(c_rgb, cv2.COLOR_RGB2BGR)
            g_bgr_flip = cv2.flip(c_bgr, 0)
            cv2.imshow('viewer', g_bgr_flip)
            if cv2.waitKey(1) & 0xFF == ord('q'):
                exit()
        else:
            self.mjViewer.render()

    # def viewer_setup(self):
    #     # self.viewer.cam.trackbodyid = 0         # id of the body to track ()
    #     self.viewer.cam.lookat[0] += 0  # x,y,z offset from the object (works if trackbodyid=-1)
    #     self.viewer.cam.lookat[1] += 0
    #     self.viewer.cam.lookat[2] += 0
    #     self.viewer.cam.distance = 0.8  # how much you "zoom in", model.stat.extent is the max limits of the arena
    #     self.viewer.cam.elevation = 0  # camera rotation around the axis in the plane going through the frame origin (if 0 you just see a line)
    #     self.viewer.cam.azimuth = 0  # camera rotation around the camera's vertical axis



from setuptools import setup, find_packages

# read the contents of your README file
from os import path

this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='yarok',
    long_description=long_description,
    long_description_content_type='text/markdown',
    packages=find_packages(),
    include_package_data=True,
    version='0.0.8',
    description='YAROK - Yet another robot framework',
    author='danfergo',
    entry_points={
        'console_scripts': [
            'yarok = yarok.__main__:main'
        ],
    },
    # scripts=['bin/lidaco'],
    author_email='danfergo@gmail.com',
    url='https://github.com/danfergo/yarok',  # use the URL to the github repo
    keywords=['yarok', 'robot', 'framework', 'mujoco'],  # arbitrary keywords
    classifiers=[],
    install_requires=[
        'mujoco_py',
    ]
)



from xml.etree.ElementTree import tostring

from mujoco_py import MjSim, load_model_from_xml

from ..components_manager import ComponentsManager
from ..mjc.interface import InterfaceMJC
from ..mjc.viewer import ViewerMJC


class PlatformMJC:
    """
        Manages the environment platform and necessary ports.
        Instantiates the corresponding interfaces.

        Manages the Mujoco Simulation Environment
    """

    def __init__(self, manager: ComponentsManager, viewer_mode='view'):
        xml = tostring(manager.xml_tree, encoding="unicode")
        self.model = load_model_from_xml(xml)
        self.sim = MjSim(self.model)
        self.manager = manager

        interfaces_mjc = {n: InterfaceMJC(n, self.sim, self.model)
                          for n, c in self.manager.components.items()
                          if 'mjc' in self.manager.data(n)['interfaces']}

        self.init_mjc_interfaces(interfaces_mjc)

        self.interfaces = {n: self.manager.data(n)['interfaces']['mjc'](interfaces_mjc[n])
                           for n, interface_mjc in interfaces_mjc.items()}

        self.viewer = ViewerMJC(self.sim, viewer_mode)

    def init_mjc_interfaces(self, interfaces_mjc):

        items = ['sensor', 'actuator', 'camera']
        for item in items:
            for idx in range(len(getattr(self.model, item + '_names'))):
                abs_name = getattr(self.model, item + '_id2name')(idx)
                col_idx = abs_name.index(':')
                component_name = abs_name[:col_idx]
                item_name = abs_name[col_idx + 1:]


                if component_name in interfaces_mjc:
                    interface = interfaces_mjc[component_name]
                    getattr(interface, item + '_name2id')[item_name] = idx
                    getattr(interface, item + 's').append(item_name)
                    interface.component_name = component_name

    def step(self):
        self.sim.step()
        x = {self.interfaces[i].step() for i in self.interfaces if hasattr(self.interfaces[i], 'step')}
        self.viewer.step()



import importlib
import os
from os import path
import os
import argparse

from .__init__ import run

import importlib.util
import sys
from enum import Enum


# adapted from https://stackoverflow.com/questions/19053707/converting-snake-case-to-lower-camel-case-lowercamelcase
def to_camel_case(snake_str):
    components = snake_str.split('_')
    # We capitalize the first letter of each component except the first one
    # with the 'title' method and join them together.
    return ''.join(x.title() for x in components)


def load(p):
    """
    Dynamically loads a module with name "name" and retrieves
    the class with the same name declared in that file.
    :param p:
    :param class_name:
    :return: loaded class.
    """

    module_name = p[p.rfind('.') + 1:]

    try:
        spec = importlib.util.spec_from_file_location(module_name, os.getcwd() + '/' + p.replace('.', '/') + ".py")
        module = importlib.util.module_from_spec(spec)
        sys.modules[module_name] = module
        spec.loader.exec_module(module)
    except FileNotFoundError:
        module = importlib.import_module(path, __package__)
    # print('>> ', p.rfind('.'), p, module_name, to_camel_case(module_name))
    return getattr(module, to_camel_case(module_name))


def main():
    """
        Converter command entry point.
    """

    parser = argparse.ArgumentParser()
    parser.add_argument('command', choices=['run'])
    parser.add_argument('component', type=str)

    # parser.add_argument('-C', '--config-file', default='config.yaml',
    #                     help='Configuration file path (default: configs.xml)')
    # parser.add_argument('-O', '--output-format', default=None,
    #                     help='Output format produced: NetCDF4, HDF5, metainfo,...')
    # parser.add_argument('-I', '--input-format', default=None,
    #                     help='Input files format as produced by the Lidar: S100, V1,...')
    # parser.add_argument('-D', '--input-path', default=None,
    #                     help='Input datasets directory path')
    # parser.add_argument('-v', '--verbose', action='store_true', default=False,
    #                     help='explain what is being done')
    parser.add_argument('-V', '--version', action='store_true', default=False,
                        help='display version information and exit')
    # parser.add_argument('--debug', action='store_true', default=False,
    #                     help='Shows internal error messages')
    # parser.add_argument('--context', default='',
    #                     help='Path to which relative paths, such as config-file or data-path, are resolved.')

    args = parser.parse_args()

    # Logger.set_args(args)
    # Logger.header()

    if not args.version:
        # opts = parser.parse_args()
        # args_dict = vars(args)
        # args_dict.pop('verbose')
        # args_dict.pop('version')
        # args_dict.pop('debug')

        # print('XXXXXXXXXXXXXXXXXXXx', args.command)
        # dir_path = os.path.dirname(os.path.realpath(__file__))

        run(load(args.component), {
            'platform_args': {
                'viewer_mode': 'view'
            }
        })

        # builder = Builder(**args_dict)
        # builder.build()

    else:
        print('version 0.0.1')
        # Logger.log('about')



from yarok.components_manager import component


class Robotiq2f85MJCInterface:

    def __init__(self, interface):
        self.interface = interface

    def move(selfs, a):
        print(a)


@component(
    interface_mj=Robotiq2f85MJCInterface
)
class robotiq_2f85:

    def __init__(self):
        pass

    def move(self, a):
        pass

    # def move(self, ):
    def close(self):
        self.move(0)
